#
#   CheMPS2: a spin-adapted implementation of DMRG for ab initio quantum chemistry
#   Copyright (C) 2013-2018 Sebastian Wouters
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import numpy as np
import math as m
import sys
import PyCheMPS2
import ctypes
import os

# Set the seed of the random number generator and cout.precision
Initializer = PyCheMPS2.PyInitialize()
Initializer.Init()

# Read in the FCIDUMP
psi4group = 7 # d2h: see chemps2/Irreps.h
filename  = b'../../tests/matrixelements/O2.CCPVDZ.FCIDUMP'
orbirreps = np.array([-1, -1], dtype=ctypes.c_int) # CheMPS2 reads it in from FCIDUMP
Ham = PyCheMPS2.PyHamiltonian( -1, psi4group, orbirreps, filename )
L = Ham.getL()

# Dump the Hamiltonian HDF5 file
Ham.save()

# Create a second Hamiltonian to read it back in
groupNum = Ham.getNGroup()
orbirreps = np.zeros([L], dtype=ctypes.c_int)
for orb in range(0, L):
    orbirreps[orb] = Ham.getOrbitalIrrep(orb)
HamLoad = PyCheMPS2.PyHamiltonian(L, groupNum, orbirreps)

# Read it back in
HamLoad.read()

# Compare both
RMSabs = 0.0
temp = Ham.getEconst()
RMSabs += temp * temp
RMS = 0.0
temp = Ham.getEconst() - HamLoad.getEconst()
RMS += temp * temp
for i1 in range(0,L):
    for i2 in range(0,L):
        temp = Ham.getTmat(i1,i2)
        RMSabs += temp * temp
        temp = Ham.getTmat(i1,i2) - HamLoad.getTmat(i1,i2)
        RMS += temp * temp
        for i3 in range(0,L):
            for i4 in range(0,L):
                temp = Ham.getVmat(i1,i2,i3,i4)
                RMSabs += temp * temp
                temp = Ham.getVmat(i1,i2,i3,i4) - HamLoad.getVmat(i1,i2,i3,i4)
                RMS += temp * temp
RMS = m.sqrt(RMS)
RMSabs = m.sqrt(RMSabs)
print("The 2-norm of all Hamiltonian matrix elements is", RMSabs)
print("The RMS difference between Ham and HamLoad is", RMS)

# Clean-up
os.system('ls -alh CheMPS2_Ham*.h5')
os.system('rm CheMPS2_Ham*.h5')
os.system('ls -alh CheMPS2_Ham*.h5')
del HamLoad
del Ham
del Initializer

# Check whether the test succeeded
if (RMS < 1e-10):
    print("================> Did test 7 succeed : yes")
else:
    print("================> Did test 7 succeed : no")




#
#   CheMPS2: a spin-adapted implementation of DMRG for ab initio quantum chemistry
#   Copyright (C) 2013-2018 Sebastian Wouters
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from libcpp.string cimport string

cdef extern from "chemps2/Hamiltonian.h" namespace "CheMPS2":
    cdef cppclass Hamiltonian:
        Hamiltonian(const int, const int, const int *) except +
        Hamiltonian(const string filename, const int psi4groupnumber) except +
        int getL()
        int getNGroup()
        int getOrbitalIrrep(const int)
        void setEconst(const double)
        void setTmat(const int, const int, const double)
        void setVmat(const int, const int, const int, const int, const double)
        double getEconst()
        double getTmat(const int, const int)
        double getVmat(const int, const int, const int, const int)
        void save()
        void read()
        void writeFCIDUMP(const string filename, const int Nelec, const int TwoS, const int TargetIrrep)




#
#   CheMPS2: a spin-adapted implementation of DMRG for ab initio quantum chemistry
#   Copyright (C) 2013-2018 Sebastian Wouters
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from libcpp cimport bool

cdef extern from "chemps2/DMRGSCFoptions.h" namespace "CheMPS2":
    cdef cppclass DMRGSCFoptions:
        DMRGSCFoptions() except +
        bool getDoDIIS()
        double getDIISGradientBranch()
        int getNumDIISVecs()
        bool getStoreDIIS()
        int getMaxIterations()
        double getGradientThreshold()
        bool getStoreUnitary()
        int getWhichActiveSpace()
        bool getDumpCorrelations()
        bool getStateAveraging()
        void setDoDIIS(const bool)
        void setDIISGradientBranch(const double)
        void setNumDIISVecs(const int)
        void setStoreDIIS(const bool)
        void setMaxIterations(const int)
        void setGradientThreshold(const double)
        void setStoreUnitary(const bool)
        void setWhichActiveSpace(const int)
        void setDumpCorrelations(const bool)
        void setStateAveraging(const bool)




#
#   CheMPS2: a spin-adapted implementation of DMRG for ab initio quantum chemistry
#   Copyright (C) 2013-2018 Sebastian Wouters
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

cimport Ham

cdef extern from "chemps2/Problem.h" namespace "CheMPS2":
    cdef cppclass Problem:
        Problem(const Ham.Hamiltonian *, const int, const int, const int) except +
        int gL()
        int gSy()
        int gIrrep(const int)
        int gTwoS()
        int gN()
        int gIrrep()
        double gEconst()
        double gMxElement(const int, const int, const int, const int)
        void setMxElement(const int, const int, const int, const int, const double)
        void SetupReorderD2h()




#
#   CheMPS2: a spin-adapted implementation of DMRG for ab initio quantum chemistry
#   Copyright (C) 2013-2018 Sebastian Wouters
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import numpy as np
import sys
import PyCheMPS2
import ctypes

# Set the seed of the random number generator and cout.precision
Initializer = PyCheMPS2.PyInitialize()
Initializer.Init()

########################################
### Square 2D Hubbard model with PBC ###
########################################

L_linear = 3                    # Linear size
L_square = L_linear * L_linear  # Number of orbitals
group = 0                       # C1 symmetry
U =  5.0                        # On-site repulsion
T = -1.0                        # Hopping term

Nelec = 9                       # Number of electrons
TwoS = 1                        # Two times the spin
Irrep = 0                       # Irrep = A (C1 symmetry)

# The Hamiltonian initializes all its matrix elements to 0.0
orbirreps = np.zeros([L_square], dtype=ctypes.c_int)
Ham = PyCheMPS2.PyHamiltonian(L_square, group, orbirreps)

# Fill with the site-basis matrix elements
for orb in range(L_square):
    Ham.setVmat(orb,orb,orb,orb,U)
for ix in range(L_linear):
    for iy in range(L_linear):
        idx1 = ix + L_linear * iy                        # This site
        idx2 = (( ix + 1 ) % L_linear) + L_linear * iy   # Right neighbour (PBC)
        idx3 = ix + L_linear * ((( iy + 1 ) % L_linear)) # Upper neighbour (PBC)
        Ham.setTmat(idx1,idx2,T)
        Ham.setTmat(idx1,idx3,T)

# Setting up the Problem
Prob = PyCheMPS2.PyProblem(Ham, TwoS, Nelec, Irrep)

# Setting up the ConvergenceScheme
# setInstruction(instruction, D, Econst, maxSweeps, noisePrefactor)
OptScheme = PyCheMPS2.PyConvergenceScheme(2) # 2 instructions
OptScheme.setInstruction(0,  500, 1e-10,  3, 0.05)
OptScheme.setInstruction(1, 1000, 1e-10, 10, 0.0 )

# Run ground state calculation
theDMRG = PyCheMPS2.PyDMRG(Prob, OptScheme)
EnergySite = theDMRG.Solve()
theDMRG.calc2DMandCorrelations()

# Clean-up DMRG
# theDMRG.deleteStoredMPS()
theDMRG.deleteStoredOperators()
del theDMRG

#################################################################################################################
### Hack: overwrite the matrix elements in momentum space (4-fold symmetry!!!) directly in the Problem object ###
#################################################################################################################
theDMRG = PyCheMPS2.PyDMRG(Prob, OptScheme) # Prob->construct_mxelem() is called in DMRG constructor
for orb1 in range(L_square):
    k1x = orb1 % L_linear
    k1y = orb1 / L_linear
    Telem1 = 2*T*(np.cos((2*np.pi*k1x)/L_linear) + np.cos((2*np.pi*k1y)/L_linear))
    for orb2 in range(L_square):
        k2x = orb2 % L_linear
        k2y = orb2 / L_linear
        Telem2 = 2*T*(np.cos((2*np.pi*k2x)/L_linear) + np.cos((2*np.pi*k2y)/L_linear))
        for orb3 in range(L_square):
            k3x = orb3 % L_linear
            k3y = orb3 / L_linear
            for orb4 in range(L_square):
                k4x = orb4 % L_linear
                k4y = orb4 / L_linear
                kx_conservation = False
                if (((k1x+k2x) % L_linear) == ((k3x+k4x) % L_linear)):
                    kx_conservation = True
                ky_conservation = False
                if (((k1y+k2y) % L_linear) == ((k3y+k4y) % L_linear)):
                    ky_conservation = True
                temp = 0.0
                if ( kx_conservation and ky_conservation ):
                    temp += U/L_square
                if (( orb1 == orb3 ) and ( orb2 == orb4 )):
                    temp += (Telem1+Telem2)/(Nelec-1)
                Prob.setMxElement(orb1,orb2,orb3,orb4,temp)
theDMRG.PreSolve() # New matrix elements require reconstruction of complementary renormalized operators
EnergyMomentum = theDMRG.Solve()
theDMRG.calc2DMandCorrelations()

# Clean-up
# theDMRG.deleteStoredMPS()
theDMRG.deleteStoredOperators()
del theDMRG
del OptScheme
del Prob
del Ham
del Initializer

# Check whether the test succeeded
if (np.fabs(EnergySite - EnergyMomentum) < 1e-8):
    print("================> Did test 9 succeed : yes")
else:
    print("================> Did test 9 succeed : no")




#
#   CheMPS2: a spin-adapted implementation of DMRG for ab initio quantum chemistry
#   Copyright (C) 2013-2018 Sebastian Wouters
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import numpy as np
cimport numpy as np
from libcpp.string cimport string
from libcpp cimport bool
np.import_array()

cimport Init
cimport ConvScheme
cimport Ham
cimport Prob
cimport Corr
cimport TwoRDM
cimport ThreeRDM
cimport DMRGsolver
cimport DMRGSCFopt
cimport DMRGSCF
cimport FullCI

cdef class PyInitialize:
    cdef Init.Initialize * thisptr
    def __cinit__(self):
        self.thisptr = new Init.Initialize()
    def __dealloc__(self):
        del self.thisptr
    def Init(self):
        self.thisptr.Init()

cdef class PyConvergenceScheme:
    cdef ConvScheme.ConvergenceScheme * thisptr
    def __cinit__(self, int nInstructions):
        self.thisptr = new ConvScheme.ConvergenceScheme(nInstructions)
    def __dealloc__(self):
        del self.thisptr
    def setInstruction(self, int instruction, int D, double Econv, int nMax, double noisePrefactor):
        self.thisptr.setInstruction(instruction, D, Econv, nMax, noisePrefactor)
    def set_instruction(self, int instruction, int D, double Econv, int nMax, double noisePrefactor, double dvdson_rtol):
        self.thisptr.set_instruction(instruction, D, Econv, nMax, noisePrefactor, dvdson_rtol)
    def getD(self, int instruction):
        return self.thisptr.get_D(instruction)
    def getEconv(self, int instruction):
        return self.thisptr.get_energy_conv(instruction)
    def getMaxSweeps(self, int instruction):
        return self.thisptr.get_max_sweeps(instruction)
    def getNoisePrefactor(self, int instruction):
        return self.thisptr.get_noise_prefactor(instruction)
    def getDavidsonRTOL(self, int instruction):
        return self.thisptr.get_dvdson_rtol(instruction)

cdef class PyHamiltonian:
    cdef Ham.Hamiltonian * thisptr
    def __cinit__(self, int Norbitals, int nGroup, np.ndarray[int, ndim=1, mode="c"] OrbIrreps not None, string filename='none'):
        if ( filename.compare('none')==0 ):
            assert OrbIrreps.flags['C_CONTIGUOUS']
            assert OrbIrreps.shape[0] == Norbitals
            self.thisptr = new Ham.Hamiltonian(Norbitals, nGroup, &OrbIrreps[0])
        else:
            self.thisptr = new Ham.Hamiltonian(filename, nGroup)
    def __dealloc__(self):
        del self.thisptr
    def getL(self):
        return self.thisptr.getL()
    def getNGroup(self):
        return self.thisptr.getNGroup()
    def getOrbitalIrrep(self, int orb):
        return self.thisptr.getOrbitalIrrep(orb)
    def setEconst(self, double value):
        self.thisptr.setEconst(value)
    def setTmat(self, int index1, int index2, double value):
        self.thisptr.setTmat(index1, index2, value)
    def setVmat(self, int index1, int index2, int index3, int index4, double value):
        self.thisptr.setVmat(index1, index2, index3, index4, value)
    def getEconst(self):
        return self.thisptr.getEconst()
    def getTmat(self, int index1, int index2):
        return self.thisptr.getTmat(index1, index2)
    def getVmat(self, int index1, int index2, int index3, int index4):
        return self.thisptr.getVmat(index1, index2, index3, index4)
    def save(self):
        self.thisptr.save()
    def read(self):
        self.thisptr.read()
    def writeFCIDUMP(self, string filename, int Nelec, int TwoS, int TargetIrrep):
        self.thisptr.writeFCIDUMP( filename, Nelec, TwoS, TargetIrrep )
        
cdef class PyProblem:
    cdef Prob.Problem * thisptr
    def __cinit__(self, PyHamiltonian Hami, int TwoS, int N, int Irrep):
        self.thisptr = new Prob.Problem(Hami.thisptr, TwoS, N, Irrep)
    def __dealloc__(self):
        del self.thisptr
    def gL(self):
        return self.thisptr.gL()
    def gSy(self):
        return self.thisptr.gSy()
    def gIrrep(self, int orb):
        return self.thisptr.gIrrep(orb)
    def gTwoS(self):
        return self.thisptr.gTwoS()
    def gN(self):
        return self.thisptr.gN()
    def gIrrep(self):
        return self.thisptr.gIrrep()
    def gEconst(self):
        return self.thisptr.gEconst()
    def gMxElement(self, int index1, int index2, int index3, int index4):
        return self.thisptr.gMxElement(index1, index2, index3, index4)
    def setMxElement(self, int index1, int index2, int index3, int index4, double value):
        self.thisptr.setMxElement(index1, index2, index3, index4, value)
    def SetupReorderD2h(self):
        self.thisptr.SetupReorderD2h()
        
cdef class PyDMRG:
    cdef DMRGsolver.DMRG * thisptr
    def __cinit__(self, PyProblem Probl, PyConvergenceScheme OptScheme, bool makechkpt=False, string tmpfolder='/tmp'):
        self.thisptr = new DMRGsolver.DMRG(Probl.thisptr, OptScheme.thisptr, makechkpt, tmpfolder)
    def __dealloc__(self):
        del self.thisptr
    def Solve(self):
        return self.thisptr.Solve()
    def PreSolve(self):
        self.thisptr.PreSolve()
    def calc2DMandCorrelations(self):
        self.thisptr.calc2DMandCorrelations()
    def calc_rdms_and_correlations(self, bool do_3rdm):
        self.thisptr.calc_rdms_and_correlations( do_3rdm )
    def deleteStoredMPS(self):
        self.thisptr.deleteStoredMPS()
    def deleteStoredOperators(self):
        self.thisptr.deleteStoredOperators()
    def activateExcitations(self, int nExcitations):
        self.thisptr.activateExcitations(nExcitations)
    def newExcitation(self, const double Eshift):
        self.thisptr.newExcitation(Eshift)
    #Access functions of the Corr.Correlations class
    def getCspin(self, int row, int col):
        return self.thisptr.getCorrelations().getCspin_HAM(row, col)
    def getCdens(self, int row, int col):
        return self.thisptr.getCorrelations().getCdens_HAM(row, col)
    def getCspinflip(self, int row, int col):
        return self.thisptr.getCorrelations().getCspinflip_HAM(row, col)
    def getCdirad(self, int row, int col):
        return self.thisptr.getCorrelations().getCdirad_HAM(row, col)
    def getMutInfo(self, int row, int col):
        return self.thisptr.getCorrelations().getMutualInformation_HAM(row, col)
    def getSingleOrbEntropy(self, int index):
        return self.thisptr.getCorrelations().SingleOrbitalEntropy_HAM(index)
    def getMutInfoDistance(self, int power):
        return self.thisptr.getCorrelations().MutualInformationDistance(power)
    def printCorrelations(self):
        self.thisptr.getCorrelations().Print()
    #Access functions of the TwoDM class
    def get2DMA(self, int i1, int i2, int i3, int i4):
        return self.thisptr.get2DM().getTwoDMA_HAM(i1, i2, i3, i4)
    def get2DMB(self, int i1, int i2, int i3, int i4):
        return self.thisptr.get2DM().getTwoDMB_HAM(i1, i2, i3, i4)
    def get2DMenergy(self):
        return self.thisptr.get2DM().energy()
    def getDoubleTrace2DMA(self):
        return self.thisptr.get2DM().trace()
    #Access functions of the ThreeDM class
    def get3DM(self, int i1, int i2, int i3, int i4, int i5, int i6):
        return self.thisptr.get3DM().get_ham_index(i1, i2, i3, i4, i5, i6)
    def Symm4RDM(self, np.ndarray[double, ndim=1, mode="c"] output not None, int ham_orb1, int ham_orb2, bool last_case):
        assert output.flags['C_CONTIGUOUS']
        self.thisptr.Symm4RDM(&output[0], ham_orb1, ham_orb2, last_case)
    def getFCIcoefficient(self, np.ndarray[int, ndim=1, mode="c"] alpha not None, np.ndarray[int, ndim=1, mode="c"] beta not None):
        assert alpha.flags['C_CONTIGUOUS']
        assert  beta.flags['C_CONTIGUOUS']
        return self.thisptr.getFCIcoefficient(&alpha[0],&beta[0])

cdef class PyDMRGSCFoptions:
    cdef DMRGSCFopt.DMRGSCFoptions * thisptr
    def __cinit__(self):
        self.thisptr = new DMRGSCFopt.DMRGSCFoptions()
    def __dealloc__(self):
        del self.thisptr
    def getDoDIIS(self):
        return self.thisptr.getDoDIIS()
    def getDIISGradientBranch(self):
        return self.thisptr.getDIISGradientBranch()
    def getNumDIISVecs(self):
        return self.thisptr.getNumDIISVecs()
    def getStoreDIIS(self):
        return self.thisptr.getStoreDIIS()
    def getMaxIterations(self):
        return self.thisptr.getMaxIterations()
    def getGradientThreshold(self):
        return self.thisptr.getGradientThreshold()
    def getStoreUnitary(self):
        return self.thisptr.getStoreUnitary()
    def getWhichActiveSpace(self):
        return self.thisptr.getWhichActiveSpace()
    def getDumpCorrelations(self):
        return self.thisptr.getDumpCorrelations()
    def getStateAveraging(self):
        return self.thisptr.getStateAveraging()
    def setDoDIIS(self, bool val):
        self.thisptr.setDoDIIS(val)
    def setDIISGradientBranch(self, double val):
        self.thisptr.setDIISGradientBranch(val)
    def setNumDIISVecs(self, int val):
        self.thisptr.setNumDIISVecs(val)
    def setStoreDIIS(self, bool val):
        self.thisptr.setStoreDIIS(val)
    def setMaxIterations(self, int val):
        self.thisptr.setMaxIterations(val)
    def setGradientThreshold(self, double val):
        self.thisptr.setGradientThreshold(val)
    def setStoreUnitary(self, bool val):
        self.thisptr.setStoreUnitary(val)
    def setWhichActiveSpace(self, int val):
        self.thisptr.setWhichActiveSpace(val)
    def setDumpCorrelations(self, bool val):
        self.thisptr.setDumpCorrelations(val)
    def setStateAveraging(self, bool val):
        self.thisptr.setStateAveraging(val)
        
cdef class PyCASSCF:
    cdef DMRGSCF.CASSCF * thisptr
    def __cinit__(self, PyHamiltonian theHam, np.ndarray[int, ndim=1, mode="c"] DOCC not None, np.ndarray[int, ndim=1, mode="c"] SOCC not None, np.ndarray[int, ndim=1, mode="c"] NOCC not None, np.ndarray[int, ndim=1, mode="c"] NDMRG not None, np.ndarray[int, ndim=1, mode="c"] NVIRT not None):
        assert  DOCC.flags['C_CONTIGUOUS']
        assert  SOCC.flags['C_CONTIGUOUS']
        assert  NOCC.flags['C_CONTIGUOUS']
        assert NDMRG.flags['C_CONTIGUOUS']
        assert NVIRT.flags['C_CONTIGUOUS']
        self.thisptr = new DMRGSCF.CASSCF(theHam.thisptr, &DOCC[0], &SOCC[0], &NOCC[0], &NDMRG[0], &NVIRT[0])
    def __dealloc__(self):
        del self.thisptr
    def solve(self, int Nel, int TwoS, int Irrep, PyConvergenceScheme OptScheme, int rootNum, PyDMRGSCFoptions theDMRGSCFopts):
        return self.thisptr.solve(Nel, TwoS, Irrep, OptScheme.thisptr, rootNum, theDMRGSCFopts.thisptr)
    def caspt2(self, int Nel, int TwoS, int Irrep, PyConvergenceScheme OptScheme, int rootNum, PyDMRGSCFoptions theDMRGSCFopts, double IPEA, double IMAG, bool PSEUDOCANONICAL):
        return self.thisptr.caspt2(Nel, TwoS, Irrep, OptScheme.thisptr, rootNum, theDMRGSCFopts.thisptr, IPEA, IMAG, PSEUDOCANONICAL)
    def solve_fci(self, int Nel, int TwoS, int Irrep, int rootNum, PyDMRGSCFoptions theDMRGSCFopts):
        return self.thisptr.solve(Nel, TwoS, Irrep, NULL, rootNum, theDMRGSCFopts.thisptr)
    def caspt2_fci(self, int Nel, int TwoS, int Irrep, int rootNum, PyDMRGSCFoptions theDMRGSCFopts, double IPEA, double IMAG, bool PSEUDOCANONICAL):
        return self.thisptr.caspt2(Nel, TwoS, Irrep, NULL, rootNum, theDMRGSCFopts.thisptr, IPEA, IMAG, PSEUDOCANONICAL)
    def deleteStoredUnitary(self):
        self.thisptr.deleteStoredUnitary()
    def deleteStoredDIIS(self):
        self.thisptr.deleteStoredDIIS()
        
cdef class PyFCI:
    cdef FullCI.FCI * thisptr
    def __cinit__(self, PyHamiltonian theHam, unsigned int Nel_up, unsigned int Nel_down, int TargetIrrep, double maxMemWorkMB=100.0, int FCIverbose=1):
        self.thisptr = new FullCI.FCI(theHam.thisptr, Nel_up, Nel_down, TargetIrrep, maxMemWorkMB, FCIverbose)
    def __dealloc__(self):
        del self.thisptr
    def getVecLength(self):
        return self.thisptr.getVecLength(0)
    def LowestEnergyDeterminant(self):
        return self.thisptr.LowestEnergyDeterminant()
    def GSDavidson(self, np.ndarray[double, ndim=1, mode="c"] inoutput not None):
        assert inoutput.flags['C_CONTIGUOUS']
        Energy = self.thisptr.GSDavidson(&inoutput[0])
        return Energy
    def CalcSpinSquared(self, np.ndarray[double, ndim=1, mode="c"] GSvector not None):
        assert GSvector.flags['C_CONTIGUOUS']
        SpinSquared = self.thisptr.CalcSpinSquared(&GSvector[0])
        return SpinSquared
    def Fill2RDM(self, np.ndarray[double, ndim=1, mode="c"] GSvector not None, np.ndarray[double, ndim=1, mode="c"] TwoRDM not None):
        assert GSvector.flags['C_CONTIGUOUS']
        assert   TwoRDM.flags['C_CONTIGUOUS']
        EnergyByContraction = self.thisptr.Fill2RDM(&GSvector[0], &TwoRDM[0])
        return EnergyByContraction
    def Fill3RDM(self, np.ndarray[double, ndim=1, mode="c"] GSvector not None, np.ndarray[double, ndim=1, mode="c"] ThreeRDM not None):
        assert GSvector.flags['C_CONTIGUOUS']
        assert ThreeRDM.flags['C_CONTIGUOUS']
        self.thisptr.Fill3RDM(&GSvector[0], &ThreeRDM[0])
    def Fill4RDM(self, np.ndarray[double, ndim=1, mode="c"] GSvector not None, np.ndarray[double, ndim=1, mode="c"] FourRDM not None):
        assert GSvector.flags['C_CONTIGUOUS']
        assert  FourRDM.flags['C_CONTIGUOUS']
        self.thisptr.Fill4RDM(&GSvector[0], &FourRDM[0])
    def Diag4RDM(self, np.ndarray[double, ndim=1, mode="c"] GSvector not None, np.ndarray[double, ndim=1, mode="c"] ThreeRDM not None, unsigned int ham_orbz, np.ndarray[double, ndim=1, mode="c"] output not None):
        assert GSvector.flags['C_CONTIGUOUS']
        assert ThreeRDM.flags['C_CONTIGUOUS']
        assert   output.flags['C_CONTIGUOUS']
        self.thisptr.Diag4RDM(&GSvector[0], &ThreeRDM[0], ham_orbz, &output[0])
    def FillRandom(self, unsigned int vecLength, np.ndarray[double, ndim=1, mode="c"] vector not None):
        assert vector.flags['C_CONTIGUOUS']
        self.thisptr.FillRandom(vecLength, &vector[0])
    def getFCIcoefficient(self, np.ndarray[int, ndim=1, mode="c"] alpha not None, np.ndarray[int, ndim=1, mode="c"] beta not None, np.ndarray[double, ndim=1, mode="c"] GSvector not None):
        assert    alpha.flags['C_CONTIGUOUS']
        assert     beta.flags['C_CONTIGUOUS']
        assert GSvector.flags['C_CONTIGUOUS']
        return self.thisptr.getFCIcoeff(&alpha[0], &beta[0], &GSvector[0])
    def RetardedGF(self, double omega, double eta, int orb_alpha, int orb_beta, bool isUp, double GSenergy, np.ndarray[double, ndim=1, mode="c"] GSvector not None, PyHamiltonian Hami):
        cdef np.ndarray[double, ndim=1, mode="c"] RePart = np.zeros([1])
        cdef np.ndarray[double, ndim=1, mode="c"] ImPart = np.zeros([1])
        assert GSvector.flags['C_CONTIGUOUS']
        assert   RePart.flags['C_CONTIGUOUS']
        assert   ImPart.flags['C_CONTIGUOUS']
        self.thisptr.RetardedGF(omega, eta, orb_alpha, orb_beta, isUp, GSenergy, &GSvector[0], Hami.thisptr, &RePart[0], &ImPart[0])
        return (RePart[0], ImPart[0])
    def RetardedGF_addition(self, double omega, double eta, int orb_alpha, int orb_beta, bool isUp, double GSenergy, np.ndarray[double, ndim=1, mode="c"] GSvector not None, PyHamiltonian Hami, np.ndarray[double, ndim=1, mode="c"] Re2RDM not None, np.ndarray[double, ndim=1, mode="c"] Im2RDM not None, np.ndarray[double, ndim=1, mode="c"] Add2RDM not None):
        cdef np.ndarray[double, ndim=1, mode="c"] RePart = np.zeros([1])
        cdef np.ndarray[double, ndim=1, mode="c"] ImPart = np.zeros([1])
        assert GSvector.flags['C_CONTIGUOUS']
        assert   RePart.flags['C_CONTIGUOUS']
        assert   ImPart.flags['C_CONTIGUOUS']
        assert   Re2RDM.flags['C_CONTIGUOUS']
        assert   Im2RDM.flags['C_CONTIGUOUS']
        assert  Add2RDM.flags['C_CONTIGUOUS']
        self.thisptr.RetardedGF_addition(omega, eta, orb_alpha, orb_beta, isUp, GSenergy, &GSvector[0], Hami.thisptr, &RePart[0], &ImPart[0], &Re2RDM[0], &Im2RDM[0], &Add2RDM[0])
        return (RePart[0], ImPart[0])
    def RetardedGF_removal(self, double omega, double eta, int orb_alpha, int orb_beta, bool isUp, double GSenergy, np.ndarray[double, ndim=1, mode="c"] GSvector not None, PyHamiltonian Hami, np.ndarray[double, ndim=1, mode="c"] Re2RDM not None, np.ndarray[double, ndim=1, mode="c"] Im2RDM not None, np.ndarray[double, ndim=1, mode="c"] Rem2RDM not None):
        cdef np.ndarray[double, ndim=1, mode="c"] RePart = np.zeros([1])
        cdef np.ndarray[double, ndim=1, mode="c"] ImPart = np.zeros([1])
        assert GSvector.flags['C_CONTIGUOUS']
        assert   RePart.flags['C_CONTIGUOUS']
        assert   ImPart.flags['C_CONTIGUOUS']
        assert   Re2RDM.flags['C_CONTIGUOUS']
        assert   Im2RDM.flags['C_CONTIGUOUS']
        assert  Rem2RDM.flags['C_CONTIGUOUS']
        self.thisptr.RetardedGF_removal(omega, eta, orb_alpha, orb_beta, isUp, GSenergy, &GSvector[0], Hami.thisptr, &RePart[0], &ImPart[0], &Re2RDM[0], &Im2RDM[0], &Rem2RDM[0])
        return (RePart[0], ImPart[0])
    def GFmatrix_add(self, double alpha, double beta, double eta, np.ndarray[int, ndim=1, mode="c"] orbsLeft not None, np.ndarray[int, ndim=1, mode="c"] orbsRight not None, bool isUp, np.ndarray[double, ndim=1, mode="c"] GSvector not None, PyHamiltonian Hami):
        cdef np.ndarray[double, ndim=1, mode="c"] RePart = np.zeros([len(orbsLeft)*len(orbsRight)])
        cdef np.ndarray[double, ndim=1, mode="c"] ImPart = np.zeros([len(orbsLeft)*len(orbsRight)])
        assert  GSvector.flags['C_CONTIGUOUS']
        assert    RePart.flags['C_CONTIGUOUS']
        assert    ImPart.flags['C_CONTIGUOUS']
        assert  orbsLeft.flags['C_CONTIGUOUS']
        assert orbsRight.flags['C_CONTIGUOUS']
        self.thisptr.GFmatrix_addition(alpha, beta, eta, &orbsLeft[0], len(orbsLeft), &orbsRight[0], len(orbsRight), isUp, &GSvector[0], Hami.thisptr, &RePart[0], &ImPart[0])
        return ( RePart, ImPart )
    def GFmatrix_rem(self, double alpha, double beta, double eta, np.ndarray[int, ndim=1, mode="c"] orbsLeft not None, np.ndarray[int, ndim=1, mode="c"] orbsRight not None, bool isUp, np.ndarray[double, ndim=1, mode="c"] GSvector not None, PyHamiltonian Hami):
        cdef np.ndarray[double, ndim=1, mode="c"] RePart = np.zeros([len(orbsLeft)*len(orbsRight)])
        cdef np.ndarray[double, ndim=1, mode="c"] ImPart = np.zeros([len(orbsLeft)*len(orbsRight)])
        assert  GSvector.flags['C_CONTIGUOUS']
        assert    RePart.flags['C_CONTIGUOUS']
        assert    ImPart.flags['C_CONTIGUOUS']
        assert  orbsLeft.flags['C_CONTIGUOUS']
        assert orbsRight.flags['C_CONTIGUOUS']
        self.thisptr.GFmatrix_removal(alpha, beta, eta, &orbsLeft[0], len(orbsLeft), &orbsRight[0], len(orbsRight), isUp, &GSvector[0], Hami.thisptr, &RePart[0], &ImPart[0])
        return ( RePart, ImPart )
    def DensityResponseGF(self, double omega, double eta, int orb_alpha, int orb_beta, double GSenergy, np.ndarray[double, ndim=1, mode="c"] GSvector not None):
        cdef np.ndarray[double, ndim=1, mode="c"] RePart = np.zeros([1])
        cdef np.ndarray[double, ndim=1, mode="c"] ImPart = np.zeros([1])
        assert GSvector.flags['C_CONTIGUOUS']
        assert   RePart.flags['C_CONTIGUOUS']
        assert   ImPart.flags['C_CONTIGUOUS']
        self.thisptr.DensityResponseGF(omega, eta, orb_alpha, orb_beta, GSenergy, &GSvector[0], &RePart[0], &ImPart[0])
        return (RePart[0], ImPart[0])
    def DensityResponseGF_forward(self, double omega, double eta, int orb_alpha, int orb_beta, double GSenergy, np.ndarray[double, ndim=1, mode="c"] GSvector not None, np.ndarray[double, ndim=1, mode="c"] Re2RDM not None, np.ndarray[double, ndim=1, mode="c"] Im2RDM not None, np.ndarray[double, ndim=1, mode="c"] Dens2RDM not None):
        cdef np.ndarray[double, ndim=1, mode="c"] RePart = np.zeros([1])
        cdef np.ndarray[double, ndim=1, mode="c"] ImPart = np.zeros([1])
        assert GSvector.flags['C_CONTIGUOUS']
        assert   RePart.flags['C_CONTIGUOUS']
        assert   ImPart.flags['C_CONTIGUOUS']
        assert   Re2RDM.flags['C_CONTIGUOUS']
        assert   Im2RDM.flags['C_CONTIGUOUS']
        assert Dens2RDM.flags['C_CONTIGUOUS']
        self.thisptr.DensityResponseGF_forward(omega, eta, orb_alpha, orb_beta, GSenergy, &GSvector[0], &RePart[0], &ImPart[0], &Re2RDM[0], &Im2RDM[0], &Dens2RDM[0])
        return (RePart[0], ImPart[0])
    def DensityResponseGF_backward(self, double omega, double eta, int orb_alpha, int orb_beta, double GSenergy, np.ndarray[double, ndim=1, mode="c"] GSvector not None, np.ndarray[double, ndim=1, mode="c"] Re2RDM not None, np.ndarray[double, ndim=1, mode="c"] Im2RDM not None, np.ndarray[double, ndim=1, mode="c"] Dens2RDM not None):
        cdef np.ndarray[double, ndim=1, mode="c"] RePart = np.zeros([1])
        cdef np.ndarray[double, ndim=1, mode="c"] ImPart = np.zeros([1])
        assert GSvector.flags['C_CONTIGUOUS']
        assert   RePart.flags['C_CONTIGUOUS']
        assert   ImPart.flags['C_CONTIGUOUS']
        assert   Re2RDM.flags['C_CONTIGUOUS']
        assert   Im2RDM.flags['C_CONTIGUOUS']
        assert Dens2RDM.flags['C_CONTIGUOUS']
        self.thisptr.DensityResponseGF_backward(omega, eta, orb_alpha, orb_beta, GSenergy, &GSvector[0], &RePart[0], &ImPart[0], &Re2RDM[0], &Im2RDM[0], &Dens2RDM[0])
        return (RePart[0], ImPart[0])





#
#   CheMPS2: a spin-adapted implementation of DMRG for ab initio quantum chemistry
#   Copyright (C) 2013-2018 Sebastian Wouters
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import numpy as np
import sys
import PyCheMPS2
import ctypes

# Set the seed of the random number generator and cout.precision
Initializer = PyCheMPS2.PyInitialize()
Initializer.Init()

# Read in the FCIDUMP
psi4group = 7 # d2h: see chemps2/Irreps.h
filename  = b'../../tests/matrixelements/N2.STO3G.FCIDUMP'
orbirreps = np.array( [ -1, -1 ], dtype=ctypes.c_int ) # CheMPS2 reads it in from FCIDUMP
Ham = PyCheMPS2.PyHamiltonian( -1, psi4group, orbirreps, filename )

# Define the symmetry sector
Nelec  = 14 # The number of electrons
TwoS   = np.array( [ 0, 2, 4, 4, 2, 2 ], dtype=ctypes.c_int ) # Two times the targeted spin
Irrep  = np.array( [ 0, 5, 0, 5, 2, 6 ], dtype=ctypes.c_int ) # The targeted irreps (Ag, B1u, Ag, ... )
E_FCI  = np.zeros( [ len( TwoS ) ] ,dtype=ctypes.c_double )
E_DMRG = np.zeros( [ len( TwoS ) ] ,dtype=ctypes.c_double )
C_dev  = np.zeros( [ len( TwoS ) ] ,dtype=ctypes.c_double )
S_FCI  = np.zeros( [ len( TwoS ) ] ,dtype=ctypes.c_double )

# Obtain a list of single species Slater determinants and their irrep
def stringlist( local_ham, num_elec ):
    arrays = []
    irreps = []
    for counter in range( 2**local_ham.getL() ):
        thestring = np.array( list( bin( counter )[ 2 : ] ) )
        theints   = map( int, thestring )
        if ( np.sum( theints ) == num_elec ):
            thearray = np.zeros( [ local_ham.getL() ], dtype=ctypes.c_int )
            thearray[ local_ham.getL() - len( theints ): ] = theints
            theirrep = 0
            for orb in range( local_ham.getL() ):
                if ( thearray[ orb ] == 1 ):
                    theirrep = theirrep ^ local_ham.getOrbitalIrrep( orb )
            arrays.append( thearray )
            irreps.append( theirrep )
    return ( arrays, irreps )

def relative_phase( L, string_up, string_down ):

    phase = 1
    for orb_down in range( L - 1 ):
        if ( string_down[ orb_down ] == 1 ):
            sum_up = np.sum( string_up[ orb_down + 1 : ] )
            if (( sum_up % 2 ) == 1 ):
                phase *= -1
    return phase

for sector in range( len( TwoS ) ):

    # Setting up the Problem
    prob = PyCheMPS2.PyProblem( Ham, TwoS[ sector ], Nelec, Irrep[ sector ] )

    # To perform DMRG, a set of convergence instructions should be provided
    opt_scheme = PyCheMPS2.PyConvergenceScheme( 2 ) # 2 instructions
    # setInstruction( counter, virtual_dimension, energy_convergence, max_sweeps, noise_prefactor, dvdson_rtol )
    opt_scheme.set_instruction( 0,  500, 1e-10,  2, 0.0, 1e-5  )
    opt_scheme.set_instruction( 1, 1000, 1e-10, 30, 0.0, 1e-10 ) # Tight convergence for accurate FCI coefficients

    # Do DMRG calculation
    dmrg_solver = PyCheMPS2.PyDMRG( prob, opt_scheme )
    E_DMRG[ sector ] = dmrg_solver.Solve()
    dmrg_solver.calc2DMandCorrelations()

    # Do FCI calculation
    nelec_up   = ( Nelec + TwoS[ sector ] ) / 2
    nelec_down = ( Nelec - TwoS[ sector ] ) / 2
    workmem_mb = 10.0
    verbose    = 1
    fci_solver = PyCheMPS2.PyFCI( Ham, nelec_up, nelec_down, Irrep[ sector ], workmem_mb, verbose )
    GSvector   = np.zeros( [ fci_solver.getVecLength() ], dtype=ctypes.c_double )
    GSvector[ fci_solver.LowestEnergyDeterminant() ] = 1.0
    E_FCI[ sector ] = fci_solver.GSDavidson( GSvector )
    S_FCI[ sector ] = fci_solver.CalcSpinSquared( GSvector )

    # Compare the FCI and DMRG determinant coefficients
    list_up,   irrep_up   = stringlist( Ham, nelec_up   )
    list_down, irrep_down = stringlist( Ham, nelec_down )
    rms_error1 = 0.0
    rms_error2 = 0.0
    for count_up in range( len( irrep_up ) ):
        for count_down in range( len( irrep_down ) ):
            if (( irrep_up[ count_up ] ^ irrep_down[ count_down ] ) == Irrep[ sector ] ):
                dmrg_coeff = dmrg_solver.getFCIcoefficient( list_up[ count_up ], list_down[ count_down ] )
                fci_coeff  =  fci_solver.getFCIcoefficient( list_up[ count_up ], list_down[ count_down ], GSvector )
                phase_diff = relative_phase( Ham.getL(), list_up[ count_up ], list_down[ count_down ] )
                temp1      = dmrg_coeff - phase_diff * fci_coeff
                temp2      = dmrg_coeff + phase_diff * fci_coeff
                rms_error1 += temp1 * temp1
                rms_error2 += temp2 * temp2
    C_dev[ sector ] = np.sqrt( min( rms_error1, rms_error2 ) ) # The global phase of the wavefunction is arbitrary, hence.
    print("RMS difference FCI and DMRG determinant coefficients =", C_dev[ sector ])

    # Clean-up
    # dmrg_solver.deleteStoredMPS()
    dmrg_solver.deleteStoredOperators()
    del fci_solver
    del dmrg_solver
    del opt_scheme
    del prob

# Clean up
del Ham
del Initializer

# Check whether the test succeeded
success = True
for sector in range( len( TwoS ) ):
    success = success and ( np.fabs( E_FCI[ sector ] - E_DMRG[ sector ] ) < 1e-8 )
    success = success and ( C_dev[ sector ] < 1e-5 )
    success = success and ( np.fabs( S_FCI[ sector ] - 0.25 * TwoS[ sector ] * ( TwoS[ sector ] + 2 ) ) < 1e-8 )
if ( success ):
    print("================> Did test 1 succeed : yes")
else:
    print("================> Did test 1 succeed : no")




import sys
sys.path.append('/home/seba') #Folder in which PySCF is installed
from pyscf import gto, scf, symm, ao2mo
from call_chemps2 import call_chemps2
import numpy as np

def fetchJK_mo( mf , DM_mo , Cmat ):

   DM_ao = np.dot( np.dot( Cmat, DM_mo ), Cmat.T )
   JK_ao = scf.hf.get_veff( mf.mol, DM_ao, 0, 0, 1 ) #Last 3 numbers: dm_last, vhf_last, hermi
   JK_mo = np.dot( np.dot( Cmat.T, JK_ao ), Cmat )
   return JK_mo

def dmrgci( mf , TwoS , Nelec , Irrep , DSU2 , Econv , MaxSweeps , NoisePrefac , frozen , active ):

    ###  Check whether the frozen and active arrays make sense  ###
    Norbs = mf.mol.nao_nr()
    assert( np.sum( frozen < Norbs ) == len(frozen) )
    assert( np.sum( active < Norbs ) == len(active) )
    assert( np.sum( frozen >= 0 ) == len(frozen) )
    assert( np.sum( active >= 0 ) == len(active) )
    mo_occ = mf.mo_occ
    for item in frozen:
        assert( mo_occ[item] == 2 )
        for item2 in active:
            assert( item != item2 )
    frozen = frozen[ frozen.argsort() ]
    active = active[ active.argsort() ]

    ###  Get orbital information  ###
    if (len(frozen) > 0):
        mo_fro = mf.mo_coeff[:, frozen]
    mo_act = mf.mo_coeff[:, active]
    Orbsym = np.array(symm.label_orb_symm(mf.mol, mf.mol.irrep_id, mf.mol.symm_orb, mf.mo_coeff))[active] # Same conventions in PySCF and CheMPS2
    for cnt in range( len( Orbsym ) ):
        # https://github.com/sunqm/pyscf/blob/master/symm/basis.py#L128
        Orbsym[ cnt ] = Orbsym[ cnt ] % 10
    
    ###  Get the JK matrix for the doubly occupied frozen orbitals  ###
    if (len(frozen) > 0):
        DM_docc = np.zeros([ Norbs ], dtype=float)
        DM_docc[ frozen ] = 2.0
        JK_docc = fetchJK_mo( mf, np.diag(DM_docc), mf.mo_coeff )
        JK_fro  = ( JK_docc[:, frozen] )[frozen, :]
        JK_act  = ( JK_docc[:, active] )[active, :]

    ###  Build the rotated MO matrix elements  ###
    Lrot  = len(active)
    OEIao = mf.mol.intor('cint1e_kin_sph') + mf.mol.intor('cint1e_nuc_sph')
    CONST = mf.mol.energy_nuc()
    OEImo = np.dot( np.dot( mo_act.T, OEIao ), mo_act )
    if (len(frozen) > 0):
        CONST += np.einsum( 'ii->', 2 * np.dot( np.dot( mo_fro.T, OEIao ), mo_fro ) + JK_fro )
        OEImo += JK_act
    #TEINT in chemical notation; also possible to pass different shapes; see http://www.sunqm.net/pyscf/tutorial.html#hf-mp2-mcscf
    TEImo = ao2mo.outcore.full_iofree(mf.mol, mo_act, compact=False).reshape(Lrot,Lrot,Lrot,Lrot)

    ###  Reorder per irrep  ### --> not strictly necessary for CheMPS2, but converges faster
    idx    = Orbsym.argsort()
    OEImo  = OEImo[:,idx]
    OEImo  = OEImo[idx,:]
    TEImo  = TEImo[:,:,:,idx]
    TEImo  = TEImo[:,:,idx,:]
    TEImo  = TEImo[:,idx,:,:]
    TEImo  = TEImo[idx,:,:,:]
    Orbsym = Orbsym[idx]

    ###  Do DMRG calculation with CheMPS2  ###
    EnergyDMRG, TwoRDM = call_chemps2( Lrot, mf.mol.groupname, Orbsym, CONST, OEImo, TEImo, TwoS, Nelec - 2*len(frozen), Irrep, DSU2, Econv, MaxSweeps, NoisePrefac )

    ###  Reorder back to original ordering  ###
    idx2 = idx.argsort()
    assert( np.linalg.norm( idx[idx2] - idx2[idx] ) < 1e-10 )
    TwoRDM = TwoRDM[:,:,:,idx2]
    TwoRDM = TwoRDM[:,:,idx2,:]
    TwoRDM = TwoRDM[:,idx2,:,:]
    TwoRDM = TwoRDM[idx2,:,:,:]

    return ( EnergyDMRG , TwoRDM )




import sys
sys.path.append('/usr/lib/python2.7/site-packages')
import PyCheMPS2
from pyscf import gto, scf, ao2mo, symm, fci
import numpy as np
import ctypes
import time

##################
#   Molecule 1   #
##################
mol1 = gto.Mole()
mol1.atom = '''
   N  0.0000   0.0000    0.000000
   N  0.0000   0.0000    1.090105
           '''
mol1.basis = 'sto-3g'
mol1.symmetry = 1
mol1.charge = 0
mol1.spin = 0 #2*S; multiplicity-1
mol1.build()

##################
#   Molecule 2   #
##################
mol2 = gto.Mole()
mol2.atom = '''
   O  0.000000000  0.00000000  0.000000000
   H  0.790689766  0.00000000  0.612217330
   H -0.790689766  0.00000000  0.612217330
           '''
mol2.basis = '6-31g'
mol2.symmetry = 1
mol2.charge = 0
mol2.spin = 0 #2*S; multiplicity-1
mol2.build()

for geval in range(1):

    if ( geval == 0 ):
        mol = mol1
        psi4group = 7 #d2h
        print "####################"
        print "#   N2 in sto-3g   #"
        print "####################"
    if ( geval == 1 ):
        mol = mol2
        psi4group = 5 #c2v
        print "####################"
        print "#   H2O in 6-31g   #"
        print "####################"
    mf = scf.RHF(mol)
    mf.verbose = 3
    mf.scf()

    #############################
    #   Build the Hamiltonian   #
    #############################
    L         = mf.mol.nao_nr()
    N         = mf.mol.nelectron
    CONST     = mf.mol.energy_nuc()
    OEImo     = np.dot( np.dot( mf.mo_coeff.T, mf.mol.intor('cint1e_kin_sph') + mf.mol.intor('cint1e_nuc_sph') ), mf.mo_coeff )
    TEImo     = ao2mo.outcore.full_iofree(mf.mol, mf.mo_coeff, compact=False).reshape(L,L,L,L)
    orb2irrep = np.array(symm.label_orb_symm(mf.mol, mf.mol.irrep_id, mf.mol.symm_orb, mf.mo_coeff), dtype=ctypes.c_int)
    for cnt in range( len( orb2irrep ) ):
        orb2irrep[ cnt ] = orb2irrep[ cnt ] % 10

    #######################################
    #   PySCF FCI calculation and 4-RDM   #
    #######################################
    fci_solver = fci.FCI(mol, mf.mo_coeff)
    fci_energy, fci_vector = fci_solver.kernel()
    print "PySCF     FCI energy =", fci_energy + CONST
    pyscf_start = time.time()
    dm1, dm2, dm3, dm4 = fci.rdm.make_dm1234('FCI4pdm_kern_spin0', fci_vector, fci_vector, L, N)
    dm1, dm2, dm3, dm4 = fci.rdm.reorder_dm1234(dm1, dm2, dm3, dm4, inplace=True)
    pyscf_trace = np.einsum( 'ii->', np.einsum( 'jjkl->kl', np.einsum( 'iijklm->jklm', np.einsum( 'iijklmno->jklmno', dm4 ) ) ) )
    pyscf_dm4 = np.array( dm4, copy=True )
    pyscf_dm4 = pyscf_dm4.swapaxes( 1, 2 ) # dm4[ 1, 5, 2, 6, 3, 7, 4, 8 ] -> dm4[ 1, 2, 5, 6, 3, 7, 4, 8 ]
    pyscf_dm4 = pyscf_dm4.swapaxes( 2, 4 ) # dm4[ 1, 2, 5, 6, 3, 7, 4, 8 ] -> dm4[ 1, 2, 3, 6, 5, 7, 4, 8 ]
    pyscf_dm4 = pyscf_dm4.swapaxes( 3, 6 ) # dm4[ 1, 2, 3, 6, 5, 7, 4, 8 ] -> dm4[ 1, 2, 3, 4, 5, 7, 6, 8 ]
    pyscf_dm4 = pyscf_dm4.swapaxes( 5, 6 ) # dm4[ 1, 2, 3, 4, 5, 7, 6, 8 ] -> dm4[ 1, 2, 3, 4, 5, 6, 7, 8 ]
    pyscf_end = time.time()

    #############################
    #   PyCheMPS2 Hamiltonian   #
    #############################
    Initializer = PyCheMPS2.PyInitialize()
    Initializer.Init()
    Ham = PyCheMPS2.PyHamiltonian( L, psi4group, orb2irrep, 'none' )
    Ham.setEconst( CONST )
    for cnt1 in range( L ):
        for cnt2 in range( L ):
            irrep12 = Ham.getOrbitalIrrep( cnt1 ) ^ Ham.getOrbitalIrrep( cnt2 )
            if ( irrep12 == 0 ):
                Ham.setTmat(cnt1, cnt2, OEImo[cnt1, cnt2])
            for cnt3 in range( L ):
                for cnt4 in range( L ):
                    irrep34 = Ham.getOrbitalIrrep( cnt3 ) ^ Ham.getOrbitalIrrep( cnt4 )
                    if ( irrep12 == irrep34 ):
                        Ham.setVmat(cnt1, cnt2, cnt3, cnt4, TEImo[cnt1, cnt3, cnt2, cnt4]) #From chemist to physics notation

    ###########################################
    #   PyCheMPS2 FCI calculation and 3-RDM   #
    ###########################################
    Irrep     = 0
    Nel_up    = ( N + mol.spin ) / 2
    Nel_down  = ( N - mol.spin ) / 2
    work_mem  = 25.0
    verbosity = 0
    theFCI    = PyCheMPS2.PyFCI(Ham, Nel_up, Nel_down, Irrep, work_mem, verbosity)
    GSvector  = np.zeros([ theFCI.getVecLength() ], dtype=ctypes.c_double)
    GSvector[ theFCI.LowestEnergyDeterminant() ] = 1.0
    EnergyFCI = theFCI.GSDavidson(GSvector)
    print "PyCheMPS2 FCI energy =", EnergyFCI
    start2    = time.time()
    FourRDM   = np.zeros([ L**8 ], dtype=ctypes.c_double)
    theFCI.Fill4RDM( GSvector, FourRDM )
    FourRDM   = FourRDM.reshape(L,L,L,L,L,L,L,L)
    print "PySCF     Tr(\'dm4\')  =", pyscf_trace
    print "PyCheMPS2 Tr(4-RDM)  =", np.einsum( 'ii->', np.einsum( 'jkjl->kl', np.einsum( 'ijkilm->jklm', np.einsum( 'ijkliqrs->jklqrs', FourRDM ) ) ) )
    end2      = time.time()
    print "Time 4-RDM PySCF     =", pyscf_end - pyscf_start, "seconds."
    print "Time 4-RDM PyCheMPS2 =", end2      - start2,      "seconds."

    ##############################
    #   Compare the two 4-RDMs   #
    ##############################
    for orb1 in range(L):
        for orb2 in range(L):
            for orb3 in range(L):
                for orb4 in range(L):
                    for orb5 in range(L):
                        for orb6 in range(L):
                            for orb7 in range(L):
                                for orb8 in range(L):
                                    temp = FourRDM[ orb1, orb2, orb3, orb4, orb5, orb6, orb7, orb8 ] - pyscf_dm4[ orb1, orb2, orb3, orb4, orb5, orb6, orb7, orb8 ]
                                    if abs( temp ) > 1e-5:
                                        print "4-RDM[",orb1,",",orb2,",",orb3,",",orb4,",",orb5,",",orb6,",",orb7,",",orb8,"] diff =", temp
    print "RMS difference PySCF and PyCheMPS2 4-RDM =", np.linalg.norm( FourRDM - pyscf_dm4 )






#!/usr/bin/env
#
#   CheMPS2: a spin-adapted implementation of DMRG for ab initio quantum chemistry
#   Copyright (C) 2013-2022 Sebastian Wouters
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import os
import numpy as np
from glob import glob
from distutils.core import setup
from distutils.extension import Extension
from distutils.command.install_data import install_data
from Cython.Distutils import build_ext

def get_depends(dirname):
    result = glob('%s/*.h' % dirname)
    result += glob('%s/*.pxd' % dirname)
    return result

setup(
    name='CheMPS2',
    version='1.8.12',
    description='A spin-adapted implementation of DMRG for ab initio quantum chemistry',
    author='Sebastian Wouters',
    author_email='sebastianwouters@gmail.com',
    download_url='https://github.com/SebWouters/CheMPS2',
    license='GNU General Public License, version 2',
    cmdclass={'build_ext': build_ext},
    ext_modules=[
        Extension("PyCheMPS2",
            sources=['PyCheMPS2.pyx'],
            depends=get_depends('.'),
            libraries=['chemps2'],
            include_dirs=[np.get_include()],
            language="c++"),
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2',
        'Programming Language :: Cython',
        'Programming Language :: C++',
        'Topic :: Science/Engineering :: Molecular Science'
    ],
)





import sys
sys.path.append('/home/seba') #Folder in which PySCF is installed
from pyscf import gto, scf, ao2mo
from dmrgci import dmrgci
import numpy as np

##############################################
#   Standard PySCF RHF or ROHF calculation   #
##############################################
mol = gto.Mole()
mol.atom = '''
   O  0.000000000  0.00000000  0.000000000;
   H  0.790689766  0.00000000  0.612217330;
   H -0.790689766  0.00000000  0.612217330
           '''
mol.basis = 'cc-pvdz'
mol.symmetry = 1
mol.charge = 0
mol.spin = 0 #2*S; multiplicity-1
mol.build()
mf = scf.RHF(mol)
mf.verbose = 0
mf.scf()

#################################
#   Creating input for dmrgci   #
#################################
TwoS  = 0             # 2*S
Nelec = mol.nelectron # Total number of electrons
Irrep = 0             # Wavefunction irrep
DSU2        = np.array([   500,  1000])
Econv       = np.array([ 1e-10, 1e-10])
MaxSweeps   = np.array([     3,    10])
NoisePrefac = np.array([   0.1,   0.0])

# In the following two lines, the orbitals are specified by their R(O)HF energy ordering
frozen = np.array([ 0 ])                                # Keep 1s core of O frozen
active = np.array([ 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]) # 5 A1, 2 B1, and 4 B2 orbitals active

##########################################
#   Do dmrgci calculation with CheMPS2   #
##########################################
E_dmrgci, TwoRDM = dmrgci( mf, TwoS, Nelec, Irrep, DSU2, Econv, MaxSweeps, NoisePrefac, frozen, active )
print "Energy( dmrgci ) =", E_dmrgci




import sys
sys.path.append('/home/seba') #Folder in which PySCF is installed
from pyscf import gto, scf, ao2mo
from dmrgci import dmrgci
import numpy as np

##############################################
#   Standard PySCF RHF or ROHF calculation   #
##############################################
mol = gto.Mole()
mol.atom = '''
  N       0.0000   0.0000   0.000000000
  N       0.0000   0.0000   1.120797413
           '''
mol.basis = 'cc-pvdz'
mol.symmetry = 1
mol.charge = 0
mol.spin = 0 #2*S; multiplicity-1
mol.build()
mf = scf.RHF(mol)
mf.verbose = 0
mf.scf()

#################################
#   Creating input for dmrgci   #
#################################
TwoS  = 0             # 2*S
Nelec = mol.nelectron # Total number of electrons
Irrep = 0             # Wavefunction irrep
DSU2        = np.array([   500,  1000])
Econv       = np.array([ 1e-10, 1e-10])
MaxSweeps   = np.array([     3,    10])
NoisePrefac = np.array([   0.1,   0.0])

# In the following two lines, the orbitals are specified by their R(O)HF energy ordering
frozen = np.array([ 0, 1 ])                                  # Keep 1s core orbitals of N2 frozen
active = np.array([ 2, 4, 10, 13, 7, 8, 3, 9, 16, 17, 5, 6]) # 4 Ag, 1 B2g, 1 B3g, 4 B1u, 1 B2u, 1 B3u in active space

##########################################
#   Do dmrgci calculation with CheMPS2   #
##########################################
E_dmrgci, TwoRDM = dmrgci( mf, TwoS, Nelec, Irrep, DSU2, Econv, MaxSweeps, NoisePrefac, frozen, active )
print "Energy( dmrgci ) =", E_dmrgci




import sys
sys.path.append('/usr/lib/python2.7/site-packages') # Folder which contains PyCheMPS2.so
import PyCheMPS2
import ctypes
import numpy as np

def call_chemps2( Norbs, GroupName, Orbsym, CONST, OEINT, TEINT, TwoS, Nelec, Irrep, DSU2, Econv, MaxSweeps, NoisePrefac ):

    assert( len( DSU2 ) == len( Econv ) )
    assert( len( DSU2 ) == len( MaxSweeps) )
    assert( len( DSU2 ) == len( NoisePrefac ) )
    
    CheMPS2_groupnamemapping = {'C1' : 0, 'Ci' : 1, 'C2' : 2, 'Cs' : 3, 'D2' : 4, 'C2v': 5, 'C2h': 6, 'D2h': 7, 'Dooh': 7}
    Group = CheMPS2_groupnamemapping[ GroupName ]
    print "Point group of the molecule =",GroupName

    Initializer = PyCheMPS2.PyInitialize()
    Initializer.Init()

    Ham = PyCheMPS2.PyHamiltonian( Norbs, Group, np.array(Orbsym, dtype=ctypes.c_int) )
    Ham.setEconst( CONST )
    for orb1 in range(Norbs):
        for orb2 in range(Norbs):
            if ( Orbsym[orb1] ^ Orbsym[orb2] == 0 ):
                Ham.setTmat(orb1, orb2, OEINT[orb1, orb2])
            for orb3 in range(Norbs):
                for orb4 in range(Norbs):
                    if ( Orbsym[orb1] ^ Orbsym[orb2] ^ Orbsym[orb3] ^ Orbsym[orb4] == 0 ):
                        # CheMPS2 uses physics notation !
                        Ham.setVmat(orb1, orb2, orb3, orb4, TEINT[orb1, orb3, orb2, orb4])

    Prob = PyCheMPS2.PyProblem(Ham, TwoS, Nelec, Irrep)
    Prob.SetupReorderD2h() # Does not matter if the group is not d2h

    OptScheme = PyCheMPS2.PyConvergenceScheme( len( DSU2 ) )
    for instruction in range( len( DSU2 ) ):
        OptScheme.setInstruction(instruction, DSU2[instruction], Econv[instruction], MaxSweeps[instruction], NoisePrefac[instruction])

    theDMRG = PyCheMPS2.PyDMRG(Prob, OptScheme)
    EnergyDMRG = theDMRG.Solve()
    theDMRG.calc2DMandCorrelations()

    TwoRDM = np.zeros([Norbs, Norbs, Norbs, Norbs], dtype=ctypes.c_double)
    for orb1 in range(Norbs):
        for orb2 in range(Norbs):
            for orb3 in range(Norbs):
                for orb4 in range(Norbs):
                    # CheMPS2 uses physics notation !
                    TwoRDM[orb1, orb3, orb2, orb4] = theDMRG.get2DMA(orb1, orb2, orb3, orb4)
    
    theDMRG.deleteStoredMPS()
    theDMRG.deleteStoredOperators()
    del theDMRG
    del OptScheme
    del Prob
    del Ham
    del Initializer
    
    return ( EnergyDMRG, TwoRDM )




#
#   CheMPS2: a spin-adapted implementation of DMRG for ab initio quantum chemistry
#   Copyright (C) 2013-2018 Sebastian Wouters
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import numpy as np
import sys
import PyCheMPS2
import ctypes

# Set the seed of the random number generator and cout.precision
Initializer = PyCheMPS2.PyInitialize()
Initializer.Init()

#######################
### BCS Hamiltonian ###
#######################

eps = np.array([ -3.5, -2.5, -1.5, -0.5, 0.5, 1.5, 2.5, 3.5 ], dtype=ctypes.c_double)
L = len( eps )
g = -1.0
power = 0.0

Nelec = L # Number of fermions in the model = Number of single-particle states
TwoS = 0  # Twice the total spin
Irrep = 0 # No point group is used, Irrep should ALWAYS be zero.

'''
   Model: h_ij = delta_ij eps[i]
          v_ijkl = delta_ij delta_kl g ( eps[i] * eps[k] ) ^ {power}
          h_ijkl = v_ijkl + ( delta_ik h_jl + delta_jl h_ik ) / ( N - 1 )
          Ham = 0.5 sum_ijkl h_ijkl sum_sigma,tau a^+_{i,sigma} a^+_{j,tau} a_{l,tau} a_{k,sigma}
'''

# The Hamiltonian initializes all its matrix elements to 0.0
orbirreps = np.zeros( [ L ], dtype=ctypes.c_int )
group = 0
Ham = PyCheMPS2.PyHamiltonian( L, group, orbirreps )

# Setting up the Problem
Prob = PyCheMPS2.PyProblem( Ham, TwoS, Nelec, Irrep )

# Setting up the ConvergenceScheme
# setInstruction(instruction, D, Econst, maxSweeps, noisePrefactor)
OptScheme = PyCheMPS2.PyConvergenceScheme( 2 )
OptScheme.setInstruction( 0,  100, 1e-10, 10, 0.5 )
OptScheme.setInstruction( 1, 1000, 1e-10, 10, 0.0 )

# Run ground state calculation
theDMRG = PyCheMPS2.PyDMRG( Prob, OptScheme )
###############################################################################################
### Hack: overwrite the matrix elements with 4-fold symmetry directly in the Problem object ###
###############################################################################################
for orb1 in range( L ):
   for orb2 in range( L ):
      eri = g * ( abs( eps[ orb1 ] * eps[ orb2 ] )**power )
      oei = ( eps[ orb1 ] + eps[ orb2 ] ) / ( Nelec - 1 )
      if ( orb1 == orb2 ):
         Prob.setMxElement( orb1, orb1, orb2, orb2, eri + oei )
      else:
         Prob.setMxElement( orb1, orb1, orb2, orb2, eri )
         Prob.setMxElement( orb1, orb2, orb1, orb2, oei )
theDMRG.PreSolve() # New matrix elements require reconstruction of complementary renormalized operators
Energy = theDMRG.Solve()
theDMRG.calc2DMandCorrelations()
theDMRG.printCorrelations()

# Clean-up
# theDMRG.deleteStoredMPS()
theDMRG.deleteStoredOperators()
del theDMRG
del OptScheme
del Prob
del Ham
del Initializer

# Check whether the test succeeded
if ( np.fabs( Energy + 25.5134137600604 ) < 1e-8 ):
    print("================> Did test 12 succeed : yes")
else:
    print("================> Did test 12 succeed : no")




#
#   CheMPS2: a spin-adapted implementation of DMRG for ab initio quantum chemistry
#   Copyright (C) 2013-2018 Sebastian Wouters
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import numpy as np
import sys
import PyCheMPS2
import ctypes

# Set the seed of the random number generator and cout.precision
Initializer = PyCheMPS2.PyInitialize()
Initializer.Init()

# Read in the FCIDUMP
psi4group = 7 # d2h: see chemps2/Irreps.h
filename  = b'../../tests/matrixelements/N2.STO3G.FCIDUMP'
orbirreps = np.array([-1, -1], dtype=ctypes.c_int) # CheMPS2 reads it in from FCIDUMP
Ham = PyCheMPS2.PyHamiltonian( -1, psi4group, orbirreps, filename )

# Define the symmetry sector
TwoS = 0     # Two times the targeted spin
N = 14       # The number of electrons
Irrep = 0    # The targeted irrep

# Setting up the Problem
Prob = PyCheMPS2.PyProblem(Ham, TwoS, N, Irrep)
Prob.SetupReorderD2h()

# Setting up the ConvergenceScheme
# setInstruction(instruction, D, Econst, maxSweeps, noisePrefactor)
OptScheme = PyCheMPS2.PyConvergenceScheme(1) # 1 instruction
OptScheme.setInstruction(0, 1000, 1e-12, 100, 0.0)

# Do DMRG calculation for the ground state
theDMRG = PyCheMPS2.PyDMRG(Prob, OptScheme)
Energy0 = theDMRG.Solve()
theDMRG.calc2DMandCorrelations()

# Calculate the lowest two excited states
theDMRG.activateExcitations(10)
theDMRG.newExcitation(20.0)
Energy1 = theDMRG.Solve()
theDMRG.calc2DMandCorrelations()
theDMRG.newExcitation(20.0)
Energy2 = theDMRG.Solve()
theDMRG.calc2DMandCorrelations()

# Clean-up
# theDMRG.deleteStoredMPS()
theDMRG.deleteStoredOperators()

# The order of deallocation matters!
del theDMRG
del OptScheme
del Prob
del Ham
del Initializer

# Check whether the test succeeded
OK0 = (np.fabs(Energy0 + 107.648250974014) < 1e-8)
OK1 = (np.fabs(Energy1 + 106.944757308768) < 1e-8)
OK2 = (np.fabs(Energy2 + 106.92314213886 ) < 1e-8)
if (OK0 and OK1 and OK2):
    print("================> Did test 5 succeed : yes")
else:
    print("================> Did test 5 succeed : no")




#
#   CheMPS2: a spin-adapted implementation of DMRG for ab initio quantum chemistry
#   Copyright (C) 2013-2018 Sebastian Wouters
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import numpy as np
import sys
import PyCheMPS2
import ctypes

# Set the seed of the random number generator and cout.precision
Initializer = PyCheMPS2.PyInitialize()
Initializer.Init()

# Read in the FCIDUMP
psi4group = 7 # d2h: see chemps2/Irreps.h
filename  = b'../../tests/matrixelements/N2.CCPVDZ.FCIDUMP'
orbirreps = np.array([-1, -1], dtype=ctypes.c_int) # CheMPS2 reads it in from FCIDUMP
Ham = PyCheMPS2.PyHamiltonian( -1, psi4group, orbirreps, filename )

# Define the symmetry sector
TwoS = 0     # Two times the targeted spin
N = 14       # The number of electrons
Irrep = 0    # The targeted irrep

# Define the CASSCF
DOCC  = np.array([ 3, 0, 0, 0, 0, 2, 1, 1 ], dtype=ctypes.c_int) # see N2.ccpvdz.out
SOCC  = np.array([ 0, 0, 0, 0, 0, 0, 0, 0 ], dtype=ctypes.c_int)
NOCC  = np.array([ 1, 0, 0, 0, 0, 1, 0, 0 ], dtype=ctypes.c_int)
NDMRG = np.array([ 2, 0, 1, 1, 0, 2, 1, 1 ], dtype=ctypes.c_int)
NVIRT = np.array([ 4, 1, 2, 2, 1, 4, 2, 2 ], dtype=ctypes.c_int)
theDMRGSCF = PyCheMPS2.PyCASSCF(Ham, DOCC, SOCC, NOCC, NDMRG, NVIRT)

# Setting up the ConvergenceScheme
OptScheme = PyCheMPS2.PyConvergenceScheme( 1 ) # 1 instruction
OptScheme.set_instruction( 0, 1000, 1e-8, 20, 0.0, 1e-8 )

# Setting the DMRGSCFoptions and run DMRGSCF
root_num = 1 # Ground state only
theDMRGSCFoptions = PyCheMPS2.PyDMRGSCFoptions()
theDMRGSCFoptions.setDoDIIS(True)
theDMRGSCFoptions.setWhichActiveSpace(1) # 1 means natural orbitals
theDMRGSCFoptions.setDumpCorrelations(True)
Energy = theDMRGSCF.solve(N, TwoS, Irrep, OptScheme, root_num, theDMRGSCFoptions)

# Clean-up
if theDMRGSCFoptions.getStoreUnitary():
    theDMRGSCF.deleteStoredUnitary()
if theDMRGSCFoptions.getStoreDIIS():
    theDMRGSCF.deleteStoredDIIS()

# The order of deallocation matters!
del theDMRGSCFoptions
del OptScheme
del theDMRGSCF
del Ham
del Initializer

# Check whether the test succeeded
if (np.fabs(Energy + 109.103502335253) < 1e-8):
    print("================> Did test 8 succeed : yes")
else:
    print("================> Did test 8 succeed : no")




#
#   CheMPS2: a spin-adapted implementation of DMRG for ab initio quantum chemistry
#   Copyright (C) 2013-2018 Sebastian Wouters
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

cimport Ham
cimport ConvScheme
cimport DMRGSCFopt
from libcpp cimport bool

cdef extern from "chemps2/FCI.h" namespace "CheMPS2":
    cdef cppclass FCI:
        FCI(Ham.Hamiltonian *, const unsigned int, const unsigned int, const int, const double, const int) except +
        unsigned int getVecLength(const int)
        unsigned int LowestEnergyDeterminant()
        void FillRandom(const unsigned int, double *)
        double GSDavidson(double *)
        double Fill2RDM(double *, double *)
        void Fill3RDM(double *, double *)
        void Fill4RDM(double *, double *)
        void Diag4RDM(double * vector, double * three_rdm, const unsigned int orbz, double * output)
        double CalcSpinSquared(double *)
        double getFCIcoeff(int *, int *, double *)
        void RetardedGF(const double, const double, const unsigned int, const unsigned int, const bool, const double, double *, Ham.Hamiltonian *, double *, double *)
        void RetardedGF_addition(const double, const double, const unsigned int, const unsigned int, const bool, const double, double *, Ham.Hamiltonian *, double *, double *, double *, double *, double *)
        void RetardedGF_removal(const double, const double, const unsigned int, const unsigned int, const bool, const double, double *, Ham.Hamiltonian *, double *, double *, double *, double *, double *)
        void GFmatrix_addition(const double, const double, const double, int *, const unsigned int, int *, const unsigned int, const bool, double *, Ham.Hamiltonian *, double *, double *)
        void GFmatrix_removal(const double, const double, const double, int *, const unsigned int, int *, const unsigned int, const bool, double *, Ham.Hamiltonian *, double *, double *)
        void DensityResponseGF(const double, const double, const unsigned int, const unsigned int, const double, double *, double *, double *)
        void DensityResponseGF_forward(const double, const double, const unsigned int, const unsigned int, const double, double *, double *, double *, double *, double *, double *)
        void DensityResponseGF_backward(const double, const double, const unsigned int, const unsigned int, const double, double *, double *, double *, double *, double *, double *)




#
#   CheMPS2: a spin-adapted implementation of DMRG for ab initio quantum chemistry
#   Copyright (C) 2013-2018 Sebastian Wouters
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

cdef extern from "chemps2/Initialize.h" namespace "CheMPS2":
    cdef cppclass Initialize:
        Initialize() except +
        void Init()




#
#   CheMPS2: a spin-adapted implementation of DMRG for ab initio quantum chemistry
#   Copyright (C) 2013-2018 Sebastian Wouters
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import numpy as np
import sys
import PyCheMPS2
import ctypes

# Set the seed of the random number generator and cout.precision
Initializer = PyCheMPS2.PyInitialize()
Initializer.Init()

# Read in the FCIDUMP
psi4group = 5 # c2v: see chemps2/Irreps.h
filename  = b'../../tests/matrixelements/CH4.STO3G.FCIDUMP'
orbirreps = np.array([-1, -1], dtype=ctypes.c_int) # CheMPS2 reads it in from FCIDUMP
Ham = PyCheMPS2.PyHamiltonian( -1, psi4group, orbirreps, filename )

# Define the symmetry sector
TwoS  = 0    # Two times the targeted spin
Nelec = 10   # The number of electrons
Irrep = 0    # The targeted irrep

# Setting up the Problem
Prob = PyCheMPS2.PyProblem(Ham, TwoS, Nelec, Irrep)

# Setting up the ConvergenceScheme
# setInstruction(instruction, D, Econst, maxSweeps, noisePrefactor)
OptScheme = PyCheMPS2.PyConvergenceScheme(2) # 2 instructions
OptScheme.setInstruction(0,   30, 1e-10,  3, 0.1)
OptScheme.setInstruction(1, 1000, 1e-10, 10, 0.0)

# Do DMRG calculation
theDMRG = PyCheMPS2.PyDMRG(Prob, OptScheme)
EnergyDMRG = theDMRG.Solve()
theDMRG.calc2DMandCorrelations()

# Do FCI calculation
Nel_up   = ( Nelec + TwoS ) / 2
Nel_down = ( Nelec - TwoS ) / 2
maxMemWorkMB = 10.0
FCIverbose = 1
theFCI = PyCheMPS2.PyFCI(Ham, Nel_up, Nel_down, Irrep, maxMemWorkMB, FCIverbose)
GSvector = np.zeros([ theFCI.getVecLength() ], dtype=ctypes.c_double)
GSvector[ theFCI.LowestEnergyDeterminant() ] = 1.0
EnergyFCI = theFCI.GSDavidson(GSvector)
theFCI.CalcSpinSquared(GSvector)
TwoRDM = np.zeros([ Ham.getL()**4 ], dtype=ctypes.c_double)
theFCI.Fill2RDM(GSvector, TwoRDM)
RMSerror2DM = 0.0
for orb1 in range(0, Ham.getL()):
    for orb2 in range(0, Ham.getL()):
        for orb3 in range(0, Ham.getL()):
            for orb4 in range(0, Ham.getL()):
                temp = TwoRDM[orb1 + Ham.getL()*(orb2 + Ham.getL()*(orb3 + Ham.getL()*orb4))] - theDMRG.get2DMA(orb1,orb2,orb3,orb4)
                RMSerror2DM += temp*temp
RMSerror2DM = np.sqrt(RMSerror2DM)
print("Frobenius norm of the difference of the DMRG and FCI 2-RDMs =", RMSerror2DM)

# Clean-up
# theDMRG.deleteStoredMPS()
theDMRG.deleteStoredOperators()
del theFCI
del theDMRG
del OptScheme
del Prob
del Ham
del Initializer

# Check whether the test succeeded
if ((np.fabs(EnergyDMRG - EnergyFCI) < 1e-8) and (RMSerror2DM < 1e-3)):
    print("================> Did test 3 succeed : yes")
else:
    print("================> Did test 3 succeed : no")




import sys
sys.path.append('/home/seba') #Folder in which PySCF is installed
from pyscf import gto, scf, ao2mo
from dmrgci import dmrgci
import numpy as np

##############################################
#   Standard PySCF RHF or ROHF calculation   #
##############################################
mol = gto.Mole()
mol.atom = '''
   O  0.000000000  0.00000000  0.000000000;
   H  0.790689766  0.00000000  0.612217330;
   H -0.790689766  0.00000000  0.612217330
           '''
mol.basis = '6-31g'
mol.symmetry = 1
mol.charge = 0
mol.spin = 0 #2*S; multiplicity-1
mol.build()
mf = scf.RHF(mol)
mf.verbose = 0
mf.scf()

#################################
#   Creating input for dmrgci   #
#################################
TwoS  = 0             # 2*S
Nelec = mol.nelectron # Total number of electrons
Irrep = 0             # Wavefunction irrep
DSU2        = np.array([   100,  3000])
Econv       = np.array([ 1e-10, 1e-10])
MaxSweeps   = np.array([     3,    10])
NoisePrefac = np.array([   0.1,   0.0])

# In the following two lines, the orbitals are specified by their R(O)HF energy ordering
frozen = np.array([ ])                      # FCI all electron all orbital calculation
active = np.array(range( mf.mol.nao_nr() )) # FCI all electron all orbital calculation

##########################################
#   Do dmrgci calculation with CheMPS2   #
##########################################
E_dmrgci, TwoRDM = dmrgci( mf, TwoS, Nelec, Irrep, DSU2, Econv, MaxSweeps, NoisePrefac, frozen, active )

##############################
#   Example usage of 2-RDM   #
##############################
Norbs = mf.mol.nao_nr()
OEINT = np.dot(np.dot( mf.mo_coeff.T, mf.mol.intor('cint1e_kin_sph') + mf.mol.intor('cint1e_nuc_sph') ), mf.mo_coeff )
TEINT = ao2mo.outcore.full_iofree(mf.mol, mf.mo_coeff, compact=False).reshape(Norbs,Norbs,Norbs,Norbs)
TwoRDMenergy = mf.mol.energy_nuc()
TwoRDM = np.array( TwoRDM )
OneRDM = np.einsum('ijkk->ij', TwoRDM) / ( Nelec - 1.0 )
TwoRDMenergy += np.sum( np.multiply( OneRDM , OEINT ) )
TwoRDMenergy += 0.5 * np.sum( np.multiply( TwoRDM , TEINT ) )
print "Energy( 2-RDM * Ham ) =", TwoRDMenergy
print "Energy( dmrgci )      =", E_dmrgci

